import {
	client,
	db,
	gateway,
	sessionManager,
	voiceStateManager,
} from "./api.js"

import {
	ActivityType,
	APIChatInputApplicationCommandInteraction,
	ApplicationCommandType,
	GatewayDispatchEvents,
	InteractionType,
	PresenceUpdateStatus,
	WithIntrinsicProps,
} from "@discordjs/core"
import { startCommand } from "./commands/start.js"
import { config } from "./config.js"
import { syncCommands } from "./utils/commands.js"
import { VoiceState } from "./utils/wrapper/VoiceState.js"
import { isMe } from "./utils/appid.js"
import { mergeAudio } from "./utils/merge.js"
import { readFile, rm } from "fs/promises"
import path from "path"

process.on("uncaughtException", (error) => console.error(error))
process.on("unhandledRejection", (error) => console.error(error))

if (config.sync) {
	await syncCommands()
}

client.on(GatewayDispatchEvents.GuildCreate, (args) => {
	console.log(`ADD GUILD: ${args.data.name}(${args.data.id})`)
	voiceStateManager.addGuild(args.data.id, args.data.voice_states)
})

client.on(GatewayDispatchEvents.GuildDelete, (args) => {
	voiceStateManager.removeGuild(args.data.id)
})

client.on(GatewayDispatchEvents.Ready, async (ready) => {
	console.log("ready", ready.data.user.username)

	await Promise.all(
		(
			await gateway.getShardIds()
		).map((shard) =>
			client
				.updatePresence(shard, {
					afk: false,
					since: Date.now(),
					status: PresenceUpdateStatus.Online,
					activities: [
						{
							type: ActivityType.Playing,
							name: "냥냥",
						},
					],
				})
				.then(() => {
					console.log("updated presence for shard", shard)
				})
		)
	)
})

client.on(GatewayDispatchEvents.InteractionCreate, async (args) => {
	if (args.data.type === InteractionType.ApplicationCommand) {
		if (args.data.data.type === ApplicationCommandType.ChatInput) {
			const command = args.data.data.name

			let commandFn:
				| ((
						args: WithIntrinsicProps<APIChatInputApplicationCommandInteraction>
				  ) => Promise<void>)
				| null = null

			switch (command) {
				case "start":
					commandFn = startCommand
					break
			}

			if (!commandFn) return

			await commandFn(
				args as WithIntrinsicProps<APIChatInputApplicationCommandInteraction>
			)
		}
	}
})

client.on(GatewayDispatchEvents.VoiceStateUpdate, async ({ data, api }) => {
	if (data.guild_id) {
		const adapter = sessionManager.adapters.get(data.guild_id)
		if (adapter) adapter.onVoiceStateUpdate(data)
		const guild = voiceStateManager.guilds.get(data.guild_id)
		if (guild) {
			const state = guild.get(data.user_id)
			if (state) {
				state.update(data)
			} else {
				guild.set(data.user_id, new VoiceState(data, data.guild_id))
			}

			const session = sessionManager.sessionMap.get(data.guild_id)

			if (session) {
				let counter = 0
				let hasMe = false

				for (const state of guild.values()) {
					if (state.channelId !== session.voiceChannelId) continue
					if (isMe(state.id)) {
						hasMe = true
						continue
					}
					counter++
				}

				if (counter === 0 || !hasMe) {
					session.vc.destroy()
					sessionManager.removeSession(data.guild_id)
					await api.channels.createMessage(session.textChannelId, {
						content: `녹화가 종료되었어요! 오디오 파일 합치는 중...`,
					})
					setTimeout(async () => {
						await mergeAudio(session.id, async (code) => {
							if (code === 0) {
								let succeeded = false
								try {
									await api.channels.createMessage(
										session.textChannelId,
										{
											content:
												"오디오 파일 합치기가 완료되었어요!",
											files: [
												{
													contentType: "audio/ogg",
													name: `${session.id}.ogg`,
													data: await readFile(
														path.resolve(
															"recordings",
															"final",
															`${session.id}.ogg`
														)
													),
												},
											],
										}
									)
									succeeded = true
								} catch {
									await api.channels.createMessage(
										session.textChannelId,
										{
											content: `오디오 파일 합치기가 완료되었...지만 파일 업로드에 실패했어요.\n파일 위치: ${path.join(
												"recordings",
												"final",
												`${session.id}.ogg`
											)}`,
										}
									)
								}

								if (succeeded) {
									await rm(
										path.resolve("recordings", session.id),
										{ recursive: true, force: true }
									)

									await db.session.delete({
										where: { id: session.id },
									})
								}
							} else {
								await api.channels.createMessage(
									session.textChannelId,
									{
										content: `ffmpeg 명령어 실행에 실패했어요! (exit code ${code}) ID: ${session.id}`,
									}
								)
							}
						})
					}, 1000)
				}
			}
		}
	}
})

client.on(GatewayDispatchEvents.VoiceServerUpdate, (args) => {
	const adapter = sessionManager.adapters.get(args.data.guild_id)
	if (adapter) adapter.onVoiceServerUpdate(args.data)
})

await gateway.connect()
